/*===========================================================================*/
/* Fellow Amiga Emulator                                                     */
/* Filesystem browsing wrapper for the GUI                                   */
/* Author: Petter Schau (peschau@online.no)                                  */
/* Author: Dan Sutherland (ds@chromerhino.demon.co.uk)                       */
/*                                                                           */
/* Linux Version (may work under other UNIXes that use glibc)                */
/*                                                                           */
/* This file is under the GNU Public License (GPL)                           */
/*===========================================================================*/

#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include "portable.h"
#include "defs.h"
#include "fswrap.h"


/*===========================================================================*/
/* Data                                                                      */
/*===========================================================================*/

DIR *fs_wrap_dirent_handle;
struct dirent *fs_wrap_dirent;
BOOLE fs_wrap_dirent_open;


/*===========================================================================*/
/* Unix has no drives                                                        */
/*===========================================================================*/

BOOLE fsWrapHasDrives(void) {
  return FALSE;
}


/*===========================================================================*/
/* Unix has no drives                                                        */
/*===========================================================================*/

BOOLE *fsWrapGetDriveMap(void) {
  return NULL;
}


/*===========================================================================*/
/* Returns the absolute full path for a filename                             */
/*===========================================================================*/

void fsWrapFullPath(STR *dst, STR *src) {
  getcwd(dst, FS_WRAP_MAX_PATH_LENGTH);
  if (strlen(dst) + strlen(src) <= FS_WRAP_MAX_PATH_LENGTH) {
    strcat(dst, src);
  }
}

/*===========================================================================*/
/* Fills in file attributes in point structure                               */
/* Return NULL on error                                                      */
/*===========================================================================*/

fs_navig_point *fsWrapMakePoint(STR *point) {
  struct stat mystat;
  fs_navig_point *fsnp = NULL;

  if (stat(point, &mystat) == 0) {
    fsnp = (fs_navig_point *) malloc(sizeof(fs_navig_point));
    strcpy(fsnp->name, point);
    if (mystat.st_mode & S_IFREG)
      fsnp->type = FS_NAVIG_FILE;
    else if (mystat.st_mode & S_IFDIR)
      fsnp->type = FS_NAVIG_DIR;
    else
      fsnp->type = FS_NAVIG_OTHER;
    fsnp->writeable = !!(mystat.st_mode & S_IWRITE);
    fsnp->size = mystat.st_size;
    fsnp->drive = 0;
    fsnp->relative = FALSE;
    fsnp->lnode = NULL;
  }
  return fsnp;
}


/*===========================================================================*/
/* Sets current directory                                                    */
/*===========================================================================*/

BOOLE fsWrapSetCWD(fs_navig_point *fs_point) {
  return chdir(fs_point->name) == 0;
}


/*===========================================================================*/
/* Returns current directory                                                 */
/*===========================================================================*/

fs_navig_point *fsWrapGetCWD(void) {
  STR tmpcwd[FS_WRAP_MAX_PATH_LENGTH];

  getcwd(tmpcwd, FS_WRAP_MAX_PATH_LENGTH);
  return fsWrapMakePoint(tmpcwd);
}


/*===========================================================================*/
/* Free allocated dirents                                                    */
/*===========================================================================*/

void fsWrapDirentsFree(void) {
}


/*===========================================================================*/
/* Reads information about named directory                                   */
/*===========================================================================*/

BOOLE fsWrapOpenDir(fs_navig_point *fs_point) {
  fsWrapCloseDir();
  fs_wrap_dirent_handle = opendir(fs_point->name);
  return (fs_wrap_dirent_open = (fs_wrap_dirent_handle != NULL));
}


/*===========================================================================*/
/* Returns current entry in the dirlisting, and advance the index            */
/*===========================================================================*/

fs_navig_point *fsWrapReadDir(void) {
  if (fs_wrap_dirent_open) {
    fs_wrap_dirent = readdir(fs_wrap_dirent_handle);
    return fsWrapMakePoint(fs_wrap_dirent->d_name);
  }
  return NULL;
}


/*===========================================================================*/
/* Terminates the current directory listing                                  */
/*===========================================================================*/

void fsWrapCloseDir(void) {
  if (fs_wrap_dirent_open) {
    closedir(fs_wrap_dirent_handle);
    fs_wrap_dirent_open = FALSE;
  }
}


/*===========================================================================*/
/* Module startup                                                            */
/*===========================================================================*/

void fsWrapStartup(void) {
  fs_wrap_dirent_open = FALSE;
}


/*===========================================================================*/
/* Module shutdown                                                           */
/*===========================================================================*/

void fsWrapShutdown(void) {
  fsWrapCloseDir();
}
