/*
 * joydrv.c - Joystick driver for XFellow.
 * Author: Dan Sutherland <ds@chromerhino.demon.co.uk>. GPL.
 *
 * This file doesn't do much since joystick events are all handled in the
 * input thread (confusingly contained in kbddrv.c). All we do here is enable
 * and disable joystick event handling on emu start/stop.
 *
 * TODO:
 * - only enable events if joystick is detected and enabled by the user
 */

#include <SDL/SDL.h>

#include "portable.h"
#include "defs.h"
#include "fellow.h"

void joyDrvHardReset(void)
{

}

void joyDrvEmulationStart(void)
{
	SDL_JoystickOpen(0);
	fprintf(stderr, "Joystick 0: %s\n", SDL_JoystickName(0));
	SDL_JoystickEventState(SDL_ENABLE);
}

void joyDrvEmulationStop(void)
{
	SDL_JoystickEventState(FALSE); /* according to the docs I should pass SDL_DISABLE, but
					this is not declared in the headers so I pass FALSE */
}

void joyDrvStartup(void)
{
	STR ls[256];

	if(! SDL_InitSubSystem(SDL_INIT_JOYSTICK)) {
		fellowAddLog("joyDrvStartup(): joystick subsystem failed to initialise\n");
	} else {
		sprintf(ls, "joyDrvStartup(): detected %i joystick(s)\n", SDL_NumJoysticks());
		fellowAddLog(ls);
	}
}

void joyDrvShutdown(void)
{
	fellowAddLog("joyDrvShutdown(): noop\n");
}

